/* Threading include file to support various threads versions.
 * Support for:
 *
 *   - Pthreads (final draft = IEEE Std. 1003.1c-1995,
 *               _POSIX_VERSION == 199506L)
 *       Uncomment the line "#define PTHREADS".
 *       Examples: AIX 4.1.x, AIX 4.2.x.
 *
 *   - Old pthreads (create/destroy instead of destroy/delete)
 *       e.g.: AIX pthreads 1.13 (AIX 3.2.4)
 *     CAREFUL: this won't work right now, some &'s must be removed
 *
 *   - Solaris threads
 *       Uncomment the line "#define SOLARIS_THREADS".
 *       Examples: SunOS 5.4.
 */

#ifndef __THREADING_H__
#define __THREADING_H__

/*#define PTHREADS*/
/*#define OLD_PTHREADS*/
#define SOLARIS_THREADS

#ifdef OLD_PTHREADS
#   define pthread_attr_init pthread_attr_create
#   define pthread_attr_destroy pthread_attr_delete
#   define pthread_mutexattr_init pthread_mutexattr_create
#   define pthread_mutexattr_destroy pthread_mutexattr_delete
#   define pthread_condattr_init pthread_condattr_create
#   define pthread_condattr_destroy pthread_condattr_delete
#   define pthread_key_create pthread_keycreate
#   define PTHREADS
#endif

extern void *MPII_temp_ptr;

#if defined(PTHREADS)
#   include <pthread.h>

    extern pthread_attr_t MPII_temp_attrib;
    extern pthread_mutexattr_t MPII_temp_mattrib;
    extern pthread_condattr_t MPII_temp_cattrib;
#   define threads_init() \
          do { \
             pthread_attr_init (&MPII_temp_attrib); \
             pthread_mutexattr_init (&MPII_temp_mattrib); \
             pthread_condattr_init (&MPII_temp_cattrib); \
          } while (0)

#   define Thread pthread_t
#   define spawn_thread(fun,arg,thread,newproc) \
          pthread_create (&(thread), &MPII_temp_attrib, fun, arg)
#   define exit_thread(i) pthread_exit ((void *) (i))
#   define thread_id() pthread_self ()

#   define Key pthread_key_t
#   define new_tsd(key) pthread_key_create (&(key), NULL)
#   define get_tsd(key) pthread_getspecific (key)
#   define set_tsd(key,val) pthread_setspecific (key, val)

#   define Mutex pthread_mutex_t
#   define new_mutex(mutex) pthread_mutex_init (&(mutex), &MPII_temp_mattrib)
#   define lock(mutex) pthread_mutex_lock (&(mutex))
#   define unlock(mutex) pthread_mutex_unlock (&(mutex))
#   define delete_mutex(mutex) pthread_mutex_destroy (&(mutex))

#   define Cond pthread_cond_t
#   define new_cond(cond) pthread_cond_init (&(cond), &MPII_temp_cattrib)
#   define notify(cond) pthread_cond_signal (&(cond))
#   define notify_all(cond) pthread_cond_broadcast (&(cond))
#   define wait(cond,mutex) pthread_cond_wait (&(cond), &(mutex))
#   define delete_cond(cond) pthread_cond_destroy (&(cond))

#elif defined(SOLARIS_THREADS)
#   include <thread.h>
#   include <synch.h>

#   define threads_init() 

#   define Thread thread_t
#   define spawn_thread(fun,arg,thread,newproc) \
          thr_create (NULL, MPII_Stack_size, fun, arg, \
                  ((newproc) ? THR_NEW_LWP : 0), &(thread))
#   define exit_thread(i) thr_exit ((void *) (i))
#   define thread_id() thr_self ()

#   define MPII_TSD             /* Solaris TSD is broken. */
#   define Key thread_key_t
#   define new_tsd(key) thr_keycreate (&(key), NULL)
#   define get_tsd(key) (thr_getspecific (key, &MPII_temp_ptr), MPII_temp_ptr)
#   define set_tsd(key,val) thr_setspecific (key, val)

#   define Mutex mutex_t
#   define new_mutex(mutex) mutex_init (&(mutex), USYNC_THREAD, 0)
#   define lock(mutex) mutex_lock (&(mutex))
#   define unlock(mutex) mutex_unlock (&(mutex))
#   define delete_mutex(mutex) mutex_destroy (&(mutex))

#   define Cond cond_t
#   define new_cond(cond) cond_init (&(cond), USYNC_THREAD, 0)
#   define notify(cond) cond_signal (&(cond))
#   define notify_all(cond) cond_broadcast (&(cond))
#   define wait(cond,mutex) cond_wait (&(cond), &(mutex))
#   define delete_cond(cond) cond_destroy (&(cond))

#endif

#ifdef MPII_TSD
#   undef Key
#   undef new_tsd
#   undef get_tsd
#   undef set_tsd

    typedef void **Key;
#   define new_tsd(key) MPII_New_tsd (&(key))
#   define get_tsd(key) MPII_Get_tsd (&(key))
#   define set_tsd(key,val) MPII_Set_tsd (&(key),val)
#endif

#endif
